%include "util.mac"
%include "vxdn.inc"
%include "icedump.inc"
%include "wiat.inc"
%include "win32n.inc"


%ifndef MAKEDEP

global TracerSysDynamicDeviceInit
global TracerSysDynamicDeviceExit
global TracerThreadInit
global TracerThreadNotExecuteable
global BreakIn.ignore_range
global Parse_Trace
global Parse_TraceX
global Service_Trace
global Service_TraceX
global VWIN32W32ServiceTable
global VWIN32_W32_ResumeThread
global VWIN32_W32_SuspendThread


extern SetCB
extern sdata
extern Parser.error
extern Parser.errorMsg
extern Error_V86
extern Error_PM16
extern Error_PMR0
extern ParseExpression
extern API.oGetTickCount
extern OT.OLastbutone
extern OT.OBranch
extern OT.OAll
extern OT.OTracethread
extern OT.OTraceprocess
extern OT.OBreaknew
extern OT.OBreakr0
extern IsThreadWin32


bits 32


; shamelessly ripped from sandpile.org ;-)
DR6_BT_MASK		equ	00008000h	; break due to TSS.T=1
DR6_BT_BIT		equ	15
DR6_BS_MASK		equ	00004000h	; break due to EFLAGS.TF=1
DR6_BS_BIT		equ	14
DR6_BD_MASK		equ	00002000h	; break due to DR7.GD=1
DR6_BD_BIT		equ	13
DR6_SMM_MASK		equ	00001000h	; (indicated SMM on some CPUs)
DR6_SMM_BIT		equ	12
DR6_B3_MASK		equ	00000008h	; break due to DR3 match
DR6_B3_BIT		equ	3
DR6_B2_MASK		equ	00000004h	; break due to DR2 match
DR6_B2_BIT		equ	2
DR6_B1_MASK		equ	00000002h	; break due to DR1 match
DR6_B1_BIT		equ	1
DR6_B0_MASK		equ	00000001h	; break due to DR0 match
DR6_B0_BIT		equ	0


; state information for the tracer on a per thread basis
;
; the theory goes like this: we attach a simple state machine to each thread
; we trace which tries to take care of thread's awareness of its being traced.
; in other words, we try to trick the thread into beleiving that it's not
; traced and has full control over the single stepping mechanism should it
; wish to use it.
;
; the state set is two dimensional in the first approach, one dimension tells
; whether the thread is trying to trace itself or not, and the second dimension
; maintains emulation related information (all instructions able to read/write
; EFLAGS.T and DR6.BS must be properly emulated).
;
; the inputs to the state machine allow monitoring context inquiries/changes
; and reacting on exceptions (most importantly on the single step trap).
;
; the state transition functions will simply modify the state information
; and the thread's context if necessary.

IDT_SIZE		EQU 0x60

struc TraceInfo
.State0:	resd 1		; 00
.State1:	resd 1		; 04
.EIPlow:	resd 1		; 08
.EIPhigh:	resd 1		; 0C
.TickCount:	resd 2		; 10
.lastCS:	resd 1		; 18
.lastEIP:	resd 1		; 1C
.lastSS:	resd 1		; 20
.lastESP:	resd 1		; 24
.IDTR:		resd 1		; 28
.IRETD:		resb 16		; 2C
.IDT:		resb 8*IDT_SIZE	; 3C
endstruc

ABORT			EQU 0
PASSDOWN		EQU 1
DONTPASS		EQU 2

ST0_SELFTRACEOFF	EQU 0
ST0_SELFTRACEON		EQU 1
ST0_SIZE		EQU 2

ST1_NOOP		EQU 0
ST1_PUSHF		EQU 1
ST1_POPF		EQU 2
ST1_IRET		EQU 3
ST1_SIDT		EQU 4
ST1_GETTICKCOUNT	EQU 5
ST1_SIZE		EQU 6

;ST1_INTxx		EQU 5
;ST1_INTO		EQU 6
;ST1_ICEBP		EQU 7
;ST1_SIZE		EQU 8

IN_DEBUG_BS		EQU 0
IN_EXCEPTION_IN		EQU 1
IN_EXCEPTION_OUT	EQU 2
IN_SIZE			EQU 3

;IN_DEBUG_BT		EQU 1
;IN_DEBUG_BD		EQU 2
;IN_DEBUG_DRx		EQU 3
;IN_DEBUG_Unknown	EQU 4
;IN_SIZE			EQU 5


segment _LDATA
	align 4
Transitions:
ST0_SelfTraceOff:
	;	IN_DEBUG_BS     IN_EXCEPTION_IN IN_EXCEPTION_OUT
.NONE:	dd	TrRet,          TrPassLie,      TrReinforce.no_selftrace
.PUSHF:	dd	TrRetEmuPushf,  TrPassLie,      TrReinforce.no_selftrace
.POPF:	dd	TrRetEmuPopf,   TrPassLie,      TrReinforce.no_selftrace
.IRET:	dd	TrRetEmuIret,   TrPassLie,      TrReinforce.no_selftrace
.SIDT:	dd	TrRetEmuSidt,   TrPassLie,      TrReinforce.no_selftrace
.GETTC:	dd	TrRetEmuGetTC,  TrPassLie,      TrReinforce.no_selftrace

ST0_SelfTraceOn:
.NONE:	dd	TrPass,         TrPass,         TrReinforce
.PUSHF:	dd	TrPassEmuPushf, TrPass,         TrReinforce
.POPF:	dd	TrPassEmuPopf,  TrPass,         TrReinforce
.IRET:	dd	TrPassEmuIret,  TrPass,         TrReinforce
.SIDT:	dd	TrPassEmuSidt,  TrPass,         TrReinforce
.GETTC:	dd	TrPassEmuGetTC, TrPass,         TrReinforce

Context:
istruc CONTEXT
iend


segment _LTEXT
;-------------------------------------------------------------------------------
; state transition functions
;
; EBX: VMCB
; EDI: R0TCB
; ESI: R0TCB->TDS (TranceInfo)
; EBP: Client Registers
;
; each function must preserve registers except for EAX which holds the return
; value (where used, currently in response to IN_DEBUG_BS only)
;
; only the thread's context and the statemachine is allowed to be changed
;-------------------------------------------------------------------------------
TrRet:
	mov	eax,DONTPASS
	retn

;-------------------------------------------------------------------------------
TrRetEmuPushf:
	call	GetLinearSSESP
	jc	@F

	btr	word [eax],8

@@
	mov	eax,DONTPASS
	retn

;-------------------------------------------------------------------------------
TrRetEmuPopf:
	bts	word [ebp+CRS.EFlags],8
	jnc	@F

	mov	dword [esi+TraceInfo.State0],ST0_SELFTRACEON

@@
	mov	eax,DONTPASS
	retn

;-------------------------------------------------------------------------------
TrRetEmuIret:
	bts	word [ebp+CRS.EFlags],8
	jnc	@F

	mov	dword [esi+TraceInfo.State0],ST0_SELFTRACEON

@@
	mov	eax,DONTPASS
	retn

;-------------------------------------------------------------------------------
TrRetEmuSidt:
	lea	eax,[esi+TraceInfo.IDT]
	push	eax
	mov	eax,[esi+TraceInfo.IDTR]
	pop	dword [eax+2]

	mov	eax,DONTPASS
	retn

;-------------------------------------------------------------------------------
TrRetEmuGetTC:
	call	TrEmuGetTC
	mov	eax,DONTPASS
	retn

;-------------------------------------------------------------------------------
TrPass:
	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrPassEmuPushf:
	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrPassEmuPopf:
	bts	word [ebp+CRS.EFlags],8
	jc	@F

	mov	dword [esi+TraceInfo.State0],ST0_SELFTRACEOFF

@@
	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrPassEmuIret:
	bts	word [ebp+CRS.EFlags],8
	jc	@F

	mov	dword [esi+TraceInfo.State0],ST0_SELFTRACEOFF

@@
	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrPassEmuSidt:
	lea	eax,[esi+TraceInfo.IDT]
	push	eax
	mov	eax,[esi+TraceInfo.IDTR]
	pop	dword [eax+2]

	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrPassEmuGetTC:
	call	TrEmuGetTC
	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrPassLie:
	btr	word [ebp+CRS.EFlags],8
	mov	eax,PASSDOWN
	retn

;-------------------------------------------------------------------------------
TrReinforce:
	call	BreakIn
	jc	@F

	retn

@@
	mov	dword [esi+TraceInfo.State0],ST0_SELFTRACEON
	bts	word [ebp+CRS.EFlags],8
	jc	.eval_opcode

	mov	dword [esi+TraceInfo.State0],ST0_SELFTRACEOFF
	jmp	short .eval_opcode

.no_selftrace:
	call	BreakIn
	jc	@F

	retn

@@
	bts	word [ebp+CRS.EFlags],8

.eval_opcode:
	add	dword [esi+TraceInfo.TickCount],byte 1
	adc	dword [esi+TraceInfo.TickCount+4],byte 0

	call	GetLinearCSEIP
	jnc	.find_opcode

	cmp	dword [esi+TraceInfo.State0],byte ST0_SELFTRACEON
	jz	@F

	btr	word [ebp+CRS.EFlags],8

@@
	call	TracerFree

	push	ebx

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	movzx	eax,word [ebp+CRS.CS]
	mov	ebx,[ebp+CRS.EIP]
	Trace_Out "ICEDUMP: TrReinforce: GetLinearCSEIP failed, CS:EIP: #ax:#ebx"
	debug_end

	pop	ebx
	retn

.find_opcode:
	push	ecx
	push	edx

	call	SkipPrefixes
	jb	.noop

.protect_start:
	cmp	byte [eax],0x9C
	jnz	@F

	mov	dword [esi+TraceInfo.State1],ST1_PUSHF
	jmp	short .ret

@@
	cmp	byte [eax],0x9D
	jnz	@F

	mov	dword [esi+TraceInfo.State1],ST1_POPF
	jmp	short .ret

@@
	cmp	byte [eax],0xCF
	jnz	@F

	mov	dword [esi+TraceInfo.State1],ST1_IRET
	jmp	short .ret

@@
	cmp	ecx,byte 3
	jbe	@F

	cmp	word [eax],0x010F		; SIDT and similar
	jnz	@F

	mov	dl,[eax+2]
	and	dl,00111000b
	cmp	dl,00001000b			; SIDT
	jnz	@F

	add	eax,byte 2
	sub	ecx,byte 2
	call	GetLenAndAddr
	jc	@F

	mov	dword [esi+TraceInfo.IDTR],eax
	mov	dword [esi+TraceInfo.State1],ST1_SIDT
	jmp	short .ret

@@
	cmp	word [eax+ecx-15],0xA165	; mov eax,[gs:0]
	jnz	@F

	cmp	dword [eax+ecx-15+2],byte 0
	jnz	@F

	cmp	eax,[API.oGetTickCount]		; best would be a check for GS
	jb	@F

	mov	dword [esi+TraceInfo.State1],ST1_GETTICKCOUNT
	jmp	short .ret

@@

.protect_end:

.noop:
.EH:
	mov	dword [esi+TraceInfo.State1],ST1_NOOP

.ret:
	pop	edx
	pop	ecx
	retn


;-------------------------------------------------------------------------------
; EBX: VMCB
; EDI: R0TCB
; ESI: R0TCB->TDS (TranceInfo)
; EBP: Client Registers
;-------------------------------------------------------------------------------
TrEmuGetTC:
	push	eax
	push	edx

	mov	eax,[esi+TraceInfo.TickCount]
	mov	edx,[esi+TraceInfo.TickCount+4]
	shrd	eax,edx,17			; simulate a 128 mips machine
	mov	[ebp+CRS.EAX],eax

	pop	edx
	pop	eax
	retn


;-------------------------------------------------------------------------------
; EBX: VMCB
; EBP: Client Registers
;
; EAX: linear SS:ESP
; stc on error
;-------------------------------------------------------------------------------
GetLinearSSESP:
	push	ecx
	push	edx

	movzx	eax,word [ebp+CRS.SS]

	test	byte [ebp+CRS.EFlags+2],2	; is client in V86 mode?
	jz	.PM

	movzx	edx,word [ebp+CRS.ESP]
	shl	eax,4
	add	eax,edx

	pop	edx
	pop	ecx
	clc
	retn

.PM:
	VMMCall	_SelectorMapFlat, ebx, eax, byte 0
	cmp	eax,byte -1
	jnz	@F

	pop	edx
	pop	ecx
	stc
	retn

@@
	mov	edx,[ebp+CRS.ESP]
	lar	ecx,[ebp+CRS.SS]		; is client stack 32 bit?
	bt	ecx,22
	jc	@F

	movzx	edx,dx

@@
	add	eax,edx

	pop	edx
	pop	ecx
	clc
	retn

;-------------------------------------------------------------------------------
; EBX: VMCB
; EBP: Client Registers
;
; EAX: linear SS:ESP
; stc on error
;-------------------------------------------------------------------------------
GetLinearCSEIP:
	push	ecx
	push	edx

	movzx	eax,word [ebp+CRS.CS]

	test	byte [ebp+CRS.EFlags+2],2	; is client in V86 mode?
	jz	.PM

	movzx	edx,word [ebp+CRS.EIP]
	shl	eax,4
	add	eax,edx

	pop	edx
	pop	ecx
	clc
	retn

.PM:
	VMMCall	_SelectorMapFlat, ebx, eax, byte 0
	cmp	eax,byte -1
	jnz	@F

	pop	edx
	pop	ecx
	stc
	retn

@@
	mov	edx,[ebp+CRS.EIP]
	lar	ecx,[ebp+CRS.CS]		; is client 32 bit?
	bt	ecx,22
	jc	@F

	movzx	edx,dx

@@
	add	eax,edx

	pop	edx
	pop	ecx
	clc
	retn

;-------------------------------------------------------------------------------
; G-RoM's stuff
;
; EAX: address
; ECX: max length
; EBP: Client Registers
;
; EAX: linear address (operand)
; stc on error (length exceeded)
;-------------------------------------------------------------------------------
GetLenAndAddr:
	push	ebx
	push	ecx
	push	edx
	push	edi
	mov	edi,esp		; needed 'cos of EH

.protect_start:
	movzx	edx,byte [eax]

	ror	edx,3
	shr	dl,3
	rol	edx,3
	inc	eax
	jmp	dword [edx*4+.Jumptable]

.ill:
.EH:
	stc

.ret:
	mov	esp,edi
	pop	edi
	pop	edx
	pop	ecx
	pop	ebx
	retn

.ga32AX:
	mov	eax,[ebp+CRS.EAX]
	sub	ecx,byte 1
	jmp	short .ret

.ga32BX:
	mov	eax,[ebp+CRS.EBX]
	sub	ecx,byte 1
	jmp	short .ret

.ga32CX:
	mov	eax,[ebp+CRS.ECX]
	sub	ecx,byte 1
	jmp	short .ret

.ga32DX:
	mov	eax,[ebp+CRS.EDX]
	sub	ecx,byte 1
	jmp	short .ret
.ga32DD:

	mov	eax,[eax]
	sub	ecx,byte 5		;new size
	jmp	short .ret

.ga32SI:
	mov	eax,[ebp+CRS.ESI]
	sub	ecx,byte 1
	jmp	short .ret

.ga32DI:
	mov	eax,[ebp+CRS.EDI]
	sub	ecx,byte 1
	jmp	short .ret

.ga32AXD:
	mov	eax,[eax]
	add	eax,[ebp+CRS.EAX]
	sub	ecx,byte 5
	jmp	short .ret

.ga32BXD:
	mov	eax,[eax]
	add	eax,[ebp+CRS.EBX]
	sub	ecx,byte 5
	jmp	short .ret

.ga32CXD:
	mov	eax,[eax]
	add	eax,[ebp+CRS.ECX]
	sub	ecx,byte 5
	jmp	short .ret

.ga32DXD:
	mov	eax,[eax]
	add	eax,[ebp+CRS.EDX]
	sub	ecx,byte 5
	jmp	short .ret

.ga32BPD:
	mov	eax,[eax]
	add	eax,[ebp+CRS.EBP]
	sub	ecx,byte 5
	jmp	short .ret

.ga32SID:
	mov	eax,[eax]
	add	eax,[ebp+CRS.ESI]
	sub	ecx,byte 5
	jmp	short .ret

.ga32DID:
	mov	eax,[eax]
	add	eax,[ebp+CRS.EDI]
	sub	ecx,byte 5
	jmp	.ret

.ga32AXB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.EAX]
	sub	ecx,byte 2
	jmp	.ret

.ga32BXB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.EBX]
	sub	ecx,byte 2
	jmp	.ret

.ga32CXB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.ECX]
	sub	ecx,byte 2
	jmp	.ret

.ga32DXB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.EDX]
	sub	ecx,byte 2
	jmp	.ret

.ga32SIB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.ESI]
	sub	ecx,byte 2
	jmp	.ret

.ga32DIB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.EDI]
	sub	ecx,byte 2
	jmp	.ret

.ga32BPB:
	movsx	eax,byte [eax]
	add	eax,[ebp+CRS.EBP]
	sub	ecx,byte 2
	jmp	.ret

.ga32S:
	movzx	edx,byte [eax]
	and	dl,111b			; implied clc
	call	[.getBaseS+4*edx]

.ga32Scommon:
	jc	near .ret

	mov	ebx,edx

	movzx	edx,byte [eax]
	shr	dl,3
	and	dl,111b
	call	[.getIndex+4*edx]

	push	edx
	movzx	edx,byte [eax]
	shr	dl,6
	call	[.scaleIndex+4*edx]	; stdcall

	lea	eax,[edx+ebx]
	sub	ecx,byte 1
	jmp	.ret

.ga32SB:
	movzx	edx,byte [eax]
	and	dl,111b			; implied clc
	call	[.getBaseSB+4*edx]
	jmp	short .ga32Scommon

.ga32SD:
	movzx	edx,byte [eax]
	and	dl,111b			; implied clc
	call	[.getBaseSD+4*edx]
	jmp	short .ga32Scommon

.gbAX:
.giAX:
	mov	edx,[ebp+CRS.EAX]
	retn

.gbCX:
.giCX:
	mov	edx,[ebp+CRS.ECX]
	retn

.gbDX:
.giDX:
	mov	edx,[ebp+CRS.EDX]
	retn

.gbBX:
.giBX:
	mov	edx,[ebp+CRS.EBX]
	retn

.gbSP:
.giSP:
	mov	edx,[ebp+CRS.ESP]
	retn

.gbBP:
	mov	edx,[eax+1]
	sub	ecx,byte 4
	retn


.gbBPB:
	movsx	edx,byte [eax+1]
	add	edx,[ebp+CRS.EBP]
	sub	ecx,byte 1
	retn

.gbBPD:
	mov	edx,[eax+1]
	add	edx,[ebp+CRS.EBP]
	sub	ecx,byte 4
	retn

.giBP:
	mov	edx,[ebp+CRS.EBP]
	retn

.gi:
	xor	edx,edx
	retn

.gbSI:
.giSI:
	mov	edx,[ebp+CRS.ESI]
	retn

.gbDI:
.giDI:
	mov	edx,[ebp+CRS.EDI]
	retn

.si1:
	mov	edx,[esp+4]
	retn	4

.si2:
	mov	edx,[esp+4]
	shl	edx,1
	retn	4

.si4:
	mov	edx,[esp+4]
	shl	edx,2
	retn	4

.si8:
	mov	edx,[esp+4]
	shl	edx,3
	retn	4

.protect_end:


segment _LDATA
	align 4
.getBaseS:
	dd	.gbAX
	dd	.gbCX
	dd	.gbDX
	dd	.gbBX
	dd	.gbSP
	dd	.gbBP
	dd	.gbSI
	dd	.gbDI

.getBaseSB:
	dd	.gbAX
	dd	.gbCX
	dd	.gbDX
	dd	.gbBX
	dd	.gbSP
	dd	.gbBPB
	dd	.gbSI
	dd	.gbDI

.getBaseSD:
	dd	.gbAX
	dd	.gbCX
	dd	.gbDX
	dd	.gbBX
	dd	.gbSP
	dd	.gbBPD
	dd	.gbSI
	dd	.gbDI

.getIndex:
	dd	.giAX
	dd	.giCX
	dd	.giDX
	dd	.giBX
	dd	.gi
	dd	.giBP
	dd	.giSI
	dd	.giDI

.scaleIndex:
	dd	.si1
	dd	.si2
	dd	.si4
	dd	.si8

.Jumptable:
	dd 	.ga32AX
	dd	.ga32CX
	dd	.ga32DX
	dd	.ga32BX
	dd	.ga32S
	dd	.ga32DD
	dd	.ga32SI
	dd	.ga32DI

	dd	.ga32AXB
	dd	.ga32CXB
	dd	.ga32DXB
	dd	.ga32BXB
	dd	.ga32SB
	dd	.ga32BPB
	dd	.ga32SIB
	dd	.ga32DIB

	dd	.ga32AXD
	dd	.ga32CXD
	dd	.ga32DXD
	dd	.ga32BXD
	dd	.ga32SD
	dd	.ga32BPD
	dd	.ga32SID
	dd	.ga32DID

	times	8 dd .ill


segment _LTEXT
;-------------------------------------------------------------------------------
; decide if winice should pop up or continue tracing
;
; EBX: VMCB
; EDI: R0TCB
; ESI: R0TCB->TDS (TranceInfo)
; EBP: Client Registers
;
; stc if will not break in
;-------------------------------------------------------------------------------
BreakIn:
	mov	eax,[ebp+CRS.EIP]
	cmp	eax,[esi+TraceInfo.EIPlow]
	jae	@F

	stc
	retn

@@
	cmp	eax,[esi+TraceInfo.EIPhigh]
	jbe	@F

	stc
	retn

@@
	cmp	byte [OT.OLastbutone],'D'
	jz	@F

	cmp	byte [OT.OBranch],'E'
	jz	@F

	cmp	byte [OT.OAll],'E'
	jz	@F

	push	ebx
	push	ecx
	push	edx

	mov	eax,[esi+TraceInfo.lastEIP]
	mov	ecx,[esi+TraceInfo.lastCS]
	mov	ebx,[esi+TraceInfo.lastESP]
	mov	edx,[esi+TraceInfo.lastSS]
	Trace_Out "ICEDUMP: LOG: CS:EIP: #cx:#eax, SS:ESP: #dx:#ebx, R0TCB: #edi"

	mov	eax,[esi+TraceInfo.TickCount]
	mov	edx,[esi+TraceInfo.TickCount+4]
	Trace_Out "ICEDUMP: LOG: instruction count: #edx:#eax, R0TCB: #edi"

	pop	edx
	pop	ecx
	pop	ebx

@@
.ignore_range:
; test and set internal winice flag
	pushfd
	cli
	mov	eax,[fPAGEIN_InProgress]	; test internal Winice flag
	cmp	dword [eax],byte 1
	mov	dword [eax],1			; set internal Winice flag to 1
	jnz	@F
	
	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: PAGEIN already in progress, cannot break in, R0TCB: #edi"
	debug_end

	popfd
	stc
	retn

@@
	popfd

; save client registers (winice will restore them)
	mov	eax,[pINT3_CleanupForPAGEIN]

%if WINICE_VERSION <= 0x405
	mov	eax,[eax+0xC]
%else
	%error please validate version WINICE_VERSION
%endif

	push	dword [ebp+CRS.EAX]
	pop	dword [eax]

	push	dword [ebp+CRS.EBX]
	pop	dword [eax+0x04]

	push	dword [ebp+CRS.ECX]
	pop	dword [eax+0x08]

	push	dword [ebp+CRS.EDX]
	pop	dword [eax+0x0C]

	push	dword [ebp+CRS.ESI]
	pop	dword [eax+0x10]

	push	dword [ebp+CRS.EDI]
	pop	dword [eax+0x14]

	push	dword [ebp+CRS.EBP]
	pop	dword [eax+0x18]

	push	dword [ebp+CRS.ESP]
	pop	dword [eax+0x1C]

	push	dword [ebp+CRS.EIP]
	pop	dword [eax+0x20]

	btr	word [ebp+CRS.EFlags],8
	push	dword [ebp+CRS.EFlags]
	pop	dword [eax+0x24]

	push	dword [ebp+CRS.CS]
	pop	dword [eax+0x28]

	push	dword [ebp+CRS.DS]
	pop	dword [eax+0x2C]

	push	dword [ebp+CRS.SS]
	pop	dword [eax+0x30]

	push	dword [ebp+CRS.ES]
	pop	dword [eax+0x34]

	push	dword [ebp+CRS.FS]
	pop	dword [eax+0x38]

	push	dword [ebp+CRS.GS]
	pop	dword [eax+0x3C]

	push	edx

	mov	edx,cr2
	push	edx
	pop	dword [eax+0x40]

	mov	edx,cr3
	push	edx
	pop	dword [eax+0x44]

	pop	edx

; setup final INT3
	mov	eax,.oINT3+4
	sub	eax,[wSelector_WINICE_Code]
	push	eax
	push	eax

	mov	eax,[oPAGEIN_INT3]		; set address of final INT3
	pop	dword [eax]

; set up CS:EIP
; !! this is NOT suitable for V86 mode threads !!
	pop	dword [ebp+CRS.EIP]

	mov	eax,[wSelector_WINICE_Code]	; set client CS:EIP
	movzx	eax,word [eax]
	mov     [ebp+CRS.CS],eax

	call	TracerFree

	clc
	retn

.oINT3:
	int3


;-------------------------------------------------------------------------------
; hook VWIN32 Get/Set Thread Context VxD APIs
;-------------------------------------------------------------------------------
HookGetSetThreadContext:
	GetDeviceServiceOrdinal eax, _VWIN32_Get_Thread_Context
	mov	esi,HookedGetThreadContext
	VMMCall	Hook_Device_Service
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: Hook_Get_Thread_Context failed"
	debug_end

	stc
	retn

@@
	GetDeviceServiceOrdinal eax, _VWIN32_Set_Thread_Context
	mov	esi,HookedSetThreadContext
	VMMCall	Hook_Device_Service
	jc	@F

	clc
	retn

@@
	GetDeviceServiceOrdinal eax, _VWIN32_Get_Thread_Context
	mov	esi,HookedGetThreadContext
	VMMCall	Unhook_Device_Service

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: Hook_Set_Thread_Context failed"
	debug_end

	stc
	retn


;-------------------------------------------------------------------------------
; unhook VWIN32 Get/Set Thread Context VxD APIs
;-------------------------------------------------------------------------------
UnhookGetSetThreadContext:
	GetDeviceServiceOrdinal eax, _VWIN32_Get_Thread_Context
	mov	esi,HookedGetThreadContext
	VMMCall	Unhook_Device_Service

	GetDeviceServiceOrdinal eax, _VWIN32_Set_Thread_Context
	mov	esi,HookedSetThreadContext
	VMMCall	Unhook_Device_Service

	clc
	retn


;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
	jmp	short HookedGetThreadContext
	jmp	[OrgGetThreadContext]

HookedGetThreadContext:
	push	esi
	push	edi

	push	dword [esp+16]		; CONTEXT
	push	dword [esp+16]		; R0TCB, can be NULL
	call	[OrgGetThreadContext]
	pop	edi			; oh yeah, these little undoc tricks ;-)
	pop	esi

	add	edi,[TDS]
	mov	edi,[edi]		; TraceInfo
	or	edi,edi
	jz	@F

	test	dword [esi+CONTEXT.cx_ContextFlags],CONTEXT_CONTROL & ~CONTEXT_i386
	jz	@F

	cmp	dword [edi+TraceInfo.State0],byte ST0_SELFTRACEOFF
	jnz	@F

	btr	word [esi+CONTEXT.cx_EFlags],8

@@
	pop	edi
	pop	esi
	retn


segment _LDATA
	align 4
OrgGetThreadContext: dd 0


segment _LTEXT
;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
	jmp	short HookedSetThreadContext
	jmp	[OrgSetThreadContext]

HookedSetThreadContext:
	push	eax
	push	ebx
	push	esi
	push	edi
	push	ebp

	push	dword [esp+28]		; CONTEXT
	push	dword [esp+28]		; R0TCB, can be NULL
	call	[OrgSetThreadContext]
	pop	edi
	pop	esi

	mov	esi,[TDS]
	mov	esi,[esi+edi]		; TraceInfo
	or	esi,esi
	jz	.ret

	mov	ebx,[edi+TCB_VMHandle]
	mov	ebp,[edi+TCB_ClientPtr]

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	mov	eax,[ebp+CRS.EFlags]
	Trace_Out "ICEDUMP: HookedSetThreadContext: EFLAGS: #eax"
	debug_end

	call	TrReinforce

.ret:
	pop	ebp
	pop	edi
	pop	esi
	pop	ebx
	pop	eax
	retn


segment _LDATA
	align 4
OrgSetThreadContext: dd 0


segment _LTEXT
;-------------------------------------------------------------------------------
; hook VWIN32 Get/Set Thread Context W32 APIs
;-------------------------------------------------------------------------------
HookW32GetSetThreadContext:
	push	ecx

	call	GetVWIN32W32ServiceTable
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: failed to look up VWIN32's W32 service table"
	debug_end

	pop	ecx
	stc
	retn

.error:
	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: failed to find W32GetThreadContext"
	debug_end

	pop	ecx
	stc
	retn

@@
	mov	eax,[VWIN32W32ServiceTable]
	mov	ecx,[eax]		; W32 API count
	sub	ecx,byte 4
	add	eax,byte 4		; walk the parameter count list

.next:
	add	eax,byte 8
	cmp	[eax],byte 2
	loopnz	.next
	jecxz	.error
	
	cmp	[eax+8],byte 2
	jnz	.next

	cmp	[eax+8+8],byte 5
	jnz	.next

	cmp	[eax+8+8+8],byte 5
	jnz	.next

	sub	eax,[VWIN32W32ServiceTable]
	shr	eax,3

	mov	[VWIN32_W32_SetThreadContext],eax
	dec	eax
	mov	[VWIN32_W32_GetThreadContext],eax
	add	eax,byte 6
	mov	[VWIN32_W32_ResumeThread],eax
	inc	eax
	mov	[VWIN32_W32_SuspendThread],eax

	mov	eax,[VWIN32_W32_GetThreadContext]
	shl	eax,3
	add	eax,byte 8
	add	eax,[VWIN32W32ServiceTable]
	push	dword [eax]
	pop	dword [OrgW32GetThreadContext]
	mov	dword [eax],HookedW32GetThreadContext

	mov	eax,[VWIN32_W32_SetThreadContext]
	shl	eax,3
	add	eax,byte 8
	add	eax,[VWIN32W32ServiceTable]
	push	dword [eax]
	pop	dword [OrgW32SetThreadContext]
	mov	dword [eax],HookedW32SetThreadContext

	pop	ecx
	clc
	retn


;-------------------------------------------------------------------------------
; unhook VWIN32 Get/Set Thread Context W32 APIs
;-------------------------------------------------------------------------------
UnhookW32GetSetThreadContext:
	mov	eax,[VWIN32_W32_GetThreadContext]
	shl	eax,3
	add	eax,byte 8
	add	eax,[VWIN32W32ServiceTable]
	push	dword [OrgW32GetThreadContext]
	pop	dword [eax]

	mov	eax,[VWIN32_W32_SetThreadContext]
	shl	eax,3
	add	eax,byte 8
	add	eax,[VWIN32W32ServiceTable]
	push	dword [OrgW32SetThreadContext]
	pop	dword [eax]

	clc
	retn


;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
GetVWIN32W32ServiceTable:
	push	ecx

	mov	eax,VWIN32_Device_ID
	VMMCall Get_DDB			; get Win32 Services Table pointer
	mov	eax,[ecx+DDB_Win32_Service_Table]
	mov	[VWIN32W32ServiceTable],eax
	sub	eax,byte 1		; stc if eax == 0

	pop	ecx
	retn


segment _LDATA
	align 4
VWIN32W32ServiceTable:		dd 0
VWIN32_W32_GetThreadContext:	dd 0
VWIN32_W32_SetThreadContext:	dd 0
VWIN32_W32_ResumeThread:	dd 0
VWIN32_W32_SuspendThread:	dd 0
OrgW32GetThreadContext:		dd 0
OrgW32SetThreadContext:		dd 0


segment _LTEXT
;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
HookedW32GetThreadContext:
	push	esi
	push	edi

	push	dword [esp+24]		; CONTEXT
	push	dword [esp+24]		; R0TCB, can be NULL
	push	dword [esp+24]
	push	dword [esp+24]
	call	[OrgW32GetThreadContext]

	mov	esi,[esp+24]

	test	dword [esi+CONTEXT.cx_ContextFlags],CONTEXT_CONTROL & ~CONTEXT_i386
	jz	@F

	mov	edi,[esp+20]
	test	edi,edi
	jnz	@F

	VMMCall	Get_Cur_Thread_Handle

@@
	add	edi,[TDS]
	mov	edi,[edi]		; TraceInfo
	or	edi,edi
	jz	@F

	cmp	dword [edi+TraceInfo.State0],byte ST0_SELFTRACEOFF
	jnz	@F

	btr	word [esi+CONTEXT.cx_EFlags],8

@@
	pop	edi
	pop	esi
	retn	16


;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
HookedW32SetThreadContext:
	push	eax
	push	ebx
	push	esi
	push	edi
	push	ebp

	push	dword [esp+36]		; CONTEXT
	push	dword [esp+36]		; R0TCB, can be NULL
	push	dword [esp+36]
	push	dword [esp+36]
	call	[OrgW32SetThreadContext]

	mov	edi,[esp+32]		; R0TCB
	test	edi,edi
	jnz	@F

	VMMCall	Get_Cur_Thread_Handle

@@
	mov	esi,[TDS]
	mov	esi,[esi+edi]		; TraceInfo
	or	esi,esi
	jz	.ret

	mov	ebx,[edi+TCB_VMHandle]
	mov	ebp,[edi+TCB_ClientPtr]

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	mov	eax,[ebp+CRS.EFlags]
	Trace_Out "ICEDUMP: HookedW32SetThreadContext: EFLAGS: #eax"
	debug_end

	call	TrReinforce

.ret:
	pop	ebp
	pop	edi
	pop	esi
	pop	ebx
	pop	eax
	retn	16


segment _LTEXT
;-------------------------------------------------------------------------------
; hook fault handlers which eventually call win32 SEH
;-------------------------------------------------------------------------------
HookPMFaults:
	push	ecx
	push	esi
	push	edi

	mov	ecx,(HookedPMFaults.end-HookedPMFaults)/PMFaultHook_size
	mov	edi,HookedPMFaults-PMFaultHook_size

@@
	add	edi,byte PMFaultHook_size
	movzx	eax,byte [edi+PMFaultHook.id]
	mov	esi,[edi+PMFaultHook.new]
	VMMCall	Hook_PM_Fault
	jc	@F

	loop	@B

	pop	edi
	pop	esi
	pop	ecx
	clc
	retn

@@
	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: Hook_PM_Fault failed for INT #eax"
	debug_end

	pop	edi
	pop	esi
	pop	ecx
	stc
	retn


;-------------------------------------------------------------------------------
; unhook our fault handlers
;-------------------------------------------------------------------------------
UnhookPMFaults:
	push	ecx
	push	esi
	push	edi
	clc
	pushfd

	mov	ecx,(HookedPMFaults.end-HookedPMFaults)/PMFaultHook_size
	mov	edi,HookedPMFaults-PMFaultHook_size

.next:
	add	edi,byte PMFaultHook_size
	movzx	eax,byte [edi+PMFaultHook.id]
	mov	esi,[edi+PMFaultHook.new]
	VMMCall	Unhook_PM_Fault
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: Unhook_PM_Fault failed for INT #eax"
	debug_end

	or	byte [esp],1		; stc

@@
	loop	.next

	popfd
	pop	edi
	pop	esi
	pop	ecx
	retn


struc PMFaultHook
.id:	resd 1
.new:	resd 1
.old:	resd 1
endstruc

segment _LDATA
	align 4
HookedPMFaults:
.int00:	dd 0x00, HookedPMFault_INT00, 0
.int01:	dd 0x01, HookedPMFault_INT01, 0
.int03:	dd 0x03, HookedPMFault_INT03, 0
.int04:	dd 0x04, HookedPMFault_INT04, 0
.int05:	dd 0x05, HookedPMFault_INT05, 0
.int06:	dd 0x06, HookedPMFault_INT06, 0
.int0B:	dd 0x0B, HookedPMFault_INT0B, 0
.int0C:	dd 0x0C, HookedPMFault_INT0C, 0
.int0D:	dd 0x0D, HookedPMFault_INT0D, 0
.int0E:	dd 0x0E, HookedPMFault_INT0E, 0
.end:


segment _LTEXT
;-------------------------------------------------------------------------------
; EBX: VMCB
; EBP: Client Registers
;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT00; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int00+PMFaultHook.old]

HookedPMFault_INT00:
	push	dword  [HookedPMFaults.int00+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT03; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int03+PMFaultHook.old]

HookedPMFault_INT03:
	push	dword  [HookedPMFaults.int03+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT04; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int04+PMFaultHook.old]

HookedPMFault_INT04:
	push	dword  [HookedPMFaults.int04+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT05; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int05+PMFaultHook.old]

HookedPMFault_INT05:
	push	dword  [HookedPMFaults.int05+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT06; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int06+PMFaultHook.old]

HookedPMFault_INT06:
	push	dword  [HookedPMFaults.int06+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT0B; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int0B+PMFaultHook.old]

HookedPMFault_INT0B:
	push	dword  [HookedPMFaults.int0B+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT0C; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int0C+PMFaultHook.old]

HookedPMFault_INT0C:
	push	dword  [HookedPMFaults.int0C+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT0D; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int0D+PMFaultHook.old]

HookedPMFault_INT0D:
	push	dword  [HookedPMFaults.int0D+PMFaultHook.old]
	jmp	short HookedPMFault_stub


;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT0E; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int0E+PMFaultHook.old]

HookedPMFault_INT0E:
	push	dword  [HookedPMFaults.int0E+PMFaultHook.old]


;-------------------------------------------------------------------------------
HookedPMFault_stub:
	push	eax
	push	ecx
	push	edx
	push	esi
	push	edi

	VMMCall	Get_Cur_Thread_Handle
	mov	esi,[TDS]
	mov	esi,[esi+edi]		; TraceInfo
	cmp	esi,byte 0
	jnz	@F

	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	retn				; chain to old

@@
	imul	eax,[esi+TraceInfo.State0],byte ST1_SIZE*IN_SIZE
	imul	edx,[esi+TraceInfo.State1],byte IN_SIZE
	lea	eax,[eax+edx+IN_EXCEPTION_IN]
	call	[Transitions+4*eax]

	cmp	eax,byte PASSDOWN
	jz	.passdown

	cmp	eax,byte DONTPASS
	jz	.dontpass

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: INTxxHook: error tracing thread, giving up... R0TCB: #edi"
	debug_end

	cmp	[esi+TraceInfo.State0],byte ST0_SELFTRACEON
	jz	@F

	call	StopTracing

@@
	call	TracerFree

	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	retn				; chain to old

.passdown:
	push	esi
	push	edi
	mov	eax,[esp+24]
	mov	ecx,[esp+20]
	mov	edx,[esp+16]
	mov	esi,[esp+12]
	mov	edi,[esp+8]

	call	[esp+28]		; chain to old
	pop	edi
	pop	esi

.dontpass:
	imul	eax,[esi+TraceInfo.State0],byte ST1_SIZE*IN_SIZE
	imul	edx,[esi+TraceInfo.State1],byte IN_SIZE
	lea	eax,[eax+edx+IN_EXCEPTION_OUT]
	call	[Transitions+4*eax]

	call	RecordAndLog

	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	add	esp,byte 4
	retn


;-------------------------------------------------------------------------------
; this mess simulates a HOOK_PROC
;-------------------------------------------------------------------------------
	jmp	short HookedPMFault_INT01; *MUST* assemble to EB,06 or EB,0A
	jmp	[HookedPMFaults.int01+PMFaultHook.old]

;-------------------------------------------------------------------------------
; the heart of the tracer engine
;
; EBX: VMCB
; EBP: Client Registers
;
; 1. detect traced thread
; 2. check EIP for being in the specified range
; 3. emulate certain instructions
; 4. handle SEH
;-------------------------------------------------------------------------------
HookedPMFault_INT01:
	push	eax
	push	ecx
	push	edx
	push	esi
	push	edi

	VMMCall	Get_Cur_Thread_Handle
	mov	esi,[TDS]
	mov	esi,[esi+edi]		; TraceInfo
	cmp	esi,byte 0
	jnz	@F

	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	jmp	[HookedPMFaults.int01+PMFaultHook.old]

@@
; send first input to the state machine
	imul	eax,[esi+TraceInfo.State0],byte ST1_SIZE*IN_SIZE
	imul	edx,[esi+TraceInfo.State1],byte IN_SIZE
	lea	eax,[eax+edx+IN_DEBUG_BS]
	call	[Transitions+4*eax]

	cmp	eax,byte PASSDOWN
	jz	.passdown

	cmp	eax,byte DONTPASS
	jz	.dontpass

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: INT01Hook: error tracing thread, giving up... R0TCB: #edi"
	debug_end

	cmp	[esi+TraceInfo.State0],byte ST0_SELFTRACEON
	jz	@F

	call	StopTracing

@@
	call	TracerFree

	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	jmp	[HookedPMFaults.int01+PMFaultHook.old]

.passdown:
	push	esi
	push	edi
	mov	eax,[esp+24]
	mov	ecx,[esp+20]
	mov	edx,[esp+16]
	mov	esi,[esp+12]
	mov	edi,[esp+8]

	call	[HookedPMFaults.int01+PMFaultHook.old]
	pop	edi
	pop	esi

.dontpass:
	imul	eax,[esi+TraceInfo.State0],byte ST1_SIZE*IN_SIZE
	imul	edx,[esi+TraceInfo.State1],byte IN_SIZE
	lea	eax,[eax+edx+IN_EXCEPTION_OUT]
	call	[Transitions+4*eax]

	call	RecordAndLog

	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	eax
	retn


segment _LDATA
	align 4
TDS:			dd 0


segment _LTEXT
;-------------------------------------------------------------------------------
; EDI: R0TCB
; ESI: R0TCB->TDS (TranceInfo)
; EBP: Client Registers
;-------------------------------------------------------------------------------
RecordAndLog:
	push	eax
	push	ebx
	push	ecx
	push	edx

	cmp	byte [OT.OAll],'E'
	jz	.log

	cmp	byte [OT.OBranch],'D'
	jz	.record

	VMMCall	Get_Cur_VM_Handle
	call	GetLinearCSEIP
	jc	.record

	call	SkipPrefixes
	jc	.record

	call	CanChangeFlow
	jc	.record

.log:
	mov	eax,[ebp+CRS.EIP]
	movzx	ecx,word [ebp+CRS.CS]
	mov	ebx,[ebp+CRS.ESP]
	movzx	edx,word [ebp+CRS.SS]
	Trace_Out "ICEDUMP: LOG: CS:EIP: #cx:#eax, SS:ESP: #dx:#ebx, R0TCB: #edi"

.record:
	push	dword [ebp+CRS.EIP]
	pop	dword [esi+TraceInfo.lastEIP]

	movzx	eax,word [ebp+CRS.CS]
	push	eax
	pop	dword [esi+TraceInfo.lastCS]

	push	dword [ebp+CRS.ESP]
	pop	dword [esi+TraceInfo.lastESP]

	movzx	eax,word [ebp+CRS.SS]
	push	eax
	pop	dword [esi+TraceInfo.lastSS]

@@
	pop	edx
	pop	ecx
	pop	ebx
	pop	eax
	retn


;-------------------------------------------------------------------------------
; eax: client EIP, after known prefixes
; ecx: max length for real opcode (considering the 15 byte limit)
;
; stc: if instruction cannot change execution flow
;-------------------------------------------------------------------------------
CanChangeFlow:
	push	esi
	mov	esi,eax

.protect_start:
	movzx	eax,byte [esi]
	bt	[FlowChangersXXno],eax
	jc	.no_change

	bt	[FlowChangersXXyes],eax
	jnc	.change

	dec	ecx
	jecxz	.no_change

	movzx	eax,byte [esi+1]
.protect_end:

	cmp	al,0x0F
	jnz	@F

	bt	[FlowChangers0F],eax
	jc	.no_change
	jmp	short .change

@@
	cmp	al,0xFF
	jnz	.no_change

	bt	[FlowChangers0F],eax
	jc	.no_change

.change:
.EH:
	pop	esi
	clc
	retn

.no_change:
	pop	esi
	stc
	retn


segment _LDATA
	align 4
FlowChangersXXno:
	  ;11111111111111110000000000000000
	  ;FEDCBA9876543210FEDCBA9876543210
	dd 11111111111111110111111111111111b ; 1F-00
	dd 11111111111111111111111111111111b ; 3F-20
	dd 11111111111111111111111111111111b ; 5F-40
	dd 00000000000000001111111111111111b ; 7F-60
	dd 11111011111111111111111111111111b ; 9F-80
	dd 11111111111111111111111111111111b ; BF-A0
	dd 11111111111111110000001111110011b ; DF-C0
	dd 01111111111111011111000011110000b ; FF-E0
	dd 0

FlowChangersXXyes:
	  ;11111111111111110000000000000000
	  ;FEDCBA9876543210FEDCBA9876543210
	dd 11111111111111111111111111111111b ; 1F-00
	dd 11111111111111111111111111111111b ; 3F-20
	dd 11111111111111111111111111111111b ; 5F-40
	dd 00000000000000001111111111111111b ; 7F-60
	dd 11111011111111111111111111111111b ; 9F-80
	dd 11111111111111111111111111111111b ; BF-A0
	dd 11111111111111110000001111110011b ; DF-C0
	dd 11111111111111011111000011110000b ; FF-E0
	dd 0

FlowChangers0F:
	  ;11111111111111110000000000000000
	  ;FEDCBA9876543210FEDCBA9876543210
	dd 11111111111111111110001101001111b ; 1F-00
	dd 00000000000011111111111101011111b ; 3F-20
	dd 11110011111111111111111111111111b ; 5F-40
	dd 11000000111111111100111111111111b ; 7F-60
	dd 11111111111111110000000000000000b ; 9F-80
	dd 11111100111111111111111111111111b ; BF-A0
	dd 11111111101011101111111111110111b ; DF-C0
	dd 01110111111011101111111110111111b ; FF-E0
	dd 0

FlowChangersFF:
	  ;11111111111111110000000000000000
	  ;FEDCBA9876543210FEDCBA9876543210
	dd 00000000000000001111111111111111b ; 1F-00
	dd 11111111111111110000000000000000b ; 3F-20
	dd 00000000000000001111111111111111b ; 5F-40
	dd 11111111111111110000000000000000b ; 7F-60
	dd 00000000000000001111111111111111b ; 9F-80
	dd 11111111111111110000000000000000b ; BF-A0
	dd 00000000000000001111111111111111b ; DF-C0
	dd 11111111111111110000000000000000b ; FF-E0
	dd 0


segment _LTEXT

;-------------------------------------------------------------------------------
; 1. hook PM faults
; 2. hook VWIN32's Get/Set Thread Context related APIs (VxD/W32)
; 2. allocate TDS
; 3. init TDS in all existing threads
;-------------------------------------------------------------------------------
TracerSysDynamicDeviceInit:
	push    ebx
	push	ecx
	push    edi

	call	InstallExceptionHandlers
	jb	.error

	call	HookPMFaults
	jb	.error_remove_exception_handlers

	call	HookGetSetThreadContext
	jb	.error_unhook_fault

	call	HookW32GetSetThreadContext
	jb	.error_unhook_threadcontext

	VMMCall _AllocateThreadDataSlot
	or      eax,eax
	jz	.error_unhook_w32threadcontext

	mov	[TDS],eax
	VMMCall	Get_Sys_VM_Handle
	VMMCall	Get_Initial_Thread_Handle
	mov     ebx,edi
	mov     ecx,[TDS]
	xor     eax,eax

@@
	mov	[edi+ecx],eax
	VMMCall	Get_Next_Thread_Handle
	cmp	ebx,edi
	jnz	@B

	pop	edi
	pop	ecx
	pop	ebx
	clc
	retn

.error_unhook_w32threadcontext:
	call	UnhookW32GetSetThreadContext

.error_unhook_threadcontext:
	call	UnhookGetSetThreadContext

.error_unhook_fault:
	call	UnhookPMFaults

.error_remove_exception_handlers:
	call	RemoveExceptionHandlers

.error:
	pop	edi
	pop	ecx
	pop	ebx
	stc
	retn


;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
InstallExceptionHandlers:
	push	esi

	mov	esi,EHS_SkipPrefixes
	VMMCall	Install_Exception_Handler
	jc	@F

	mov	esi,EHS_TrReinforce
	VMMCall	Install_Exception_Handler
	jc	@F

	mov	esi,EHS_CanChangeFlow
	VMMCall	Install_Exception_Handler
	jc	@F

	mov	esi,EHS_GetLenAndAddr
	VMMCall	Install_Exception_Handler
	jc	@F

	pop	esi
	clc
	retn

@@
	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: InstallExceptionHandlers: failed to install exception handler"
	debug_end

	pop	esi
	stc
	retn


;-------------------------------------------------------------------------------
;
;-------------------------------------------------------------------------------
RemoveExceptionHandlers:
	push	esi

	mov	esi,EHS_SkipPrefixes
	cmp	dword [esi+EHS_Reserved],byte 0
	jz	@F

	VMMCall	Remove_Exception_Handler
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: RemoveExceptionHandlers: failed for SkipPrefixes"
	debug_end

@@
	mov	esi,EHS_TrReinforce
	cmp	dword [esi+EHS_Reserved],byte 0
	jz	@F

	VMMCall	Remove_Exception_Handler
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: RemoveExceptionHandlers: failed for TrReinforce"
	debug_end

@@
	mov	esi,EHS_CanChangeFlow
	cmp	dword [esi+EHS_Reserved],byte 0
	jz	@F

	VMMCall	Remove_Exception_Handler
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: RemoveExceptionHandlers: failed for CanChangeFlow"
	debug_end

@@
	mov	esi,EHS_GetLenAndAddr
	cmp	dword [esi+EHS_Reserved],byte 0
	jz	@F

	VMMCall	Remove_Exception_Handler
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: RemoveExceptionHandlers: failed for GetLenAndAddr"
	debug_end

@@
	pop	esi
	clc
	retn


segment _LDATA
	align 4
EHS_SkipPrefixes:
istruc Exception_Handler_Struc
	at EHS_Reserved,	dd 0
	at EHS_Start_EIP,	dd SkipPrefixes.protect_start
	at EHS_End_EIP,		dd SkipPrefixes.protect_end
	at EHS_Handler,		dd SkipPrefixes.EH
iend

EHS_TrReinforce:
istruc Exception_Handler_Struc
	at EHS_Reserved,	dd 0
	at EHS_Start_EIP,	dd TrReinforce.protect_start
	at EHS_End_EIP,		dd TrReinforce.protect_end
	at EHS_Handler,		dd TrReinforce.EH
iend

EHS_CanChangeFlow:
istruc Exception_Handler_Struc
	at EHS_Reserved,	dd 0
	at EHS_Start_EIP,	dd CanChangeFlow.protect_start
	at EHS_End_EIP,		dd CanChangeFlow.protect_end
	at EHS_Handler,		dd CanChangeFlow.EH
iend

EHS_GetLenAndAddr:
istruc Exception_Handler_Struc
	at EHS_Reserved,	dd 0
	at EHS_Start_EIP,	dd GetLenAndAddr.protect_start
	at EHS_End_EIP,		dd GetLenAndAddr.protect_end
	at EHS_Handler,		dd GetLenAndAddr.EH
iend


segment _LTEXT
;-------------------------------------------------------------------------------
; clean up all tracer structures
;-------------------------------------------------------------------------------
TracerSysDynamicDeviceExit:
	push	ebx
	push	ecx
	push	edi

	VMMCall	Get_Sys_VM_Handle
	VMMCall	Get_Initial_Thread_Handle
	mov     ebx,edi
	mov     ecx,[TDS]

.loop:
	mov	eax,[ecx+edi]
	or	eax,eax
	jz	.next

	cmp	[eax+TraceInfo.State0],byte ST0_SELFTRACEON
	jz	@F

	call	StopTracing

@@
	call	TracerFree

.next:
	VMMCall	Get_Next_Thread_Handle
	cmp	ebx,edi
	jnz	.loop

	VMMCall	_FreeThreadDataSlot, dword [TDS]

	call	UnhookW32GetSetThreadContext

	call	UnhookGetSetThreadContext

	call	UnhookPMFaults

	pop	edi
	pop	ecx
	pop	ebx
	clc
	retn


;-------------------------------------------------------------------------------
; init TDS in new thread
; check if it belongs to a process one of whose threads is already being traced
; if so and multithread tracing is enabled, trace this guy as well
;
; EDI: R0TCB
;-------------------------------------------------------------------------------
TracerThreadInit:
	push	ebx
	push	ecx
	push	edx
	push	esi
	push	edi
	push	ebp

	mov	ecx,[TDS]
	and	dword [ecx+edi],byte 0

	mov	esi,edi				; save current R0TCB
	mov	ebp,[edi+TCB_ClientPtr]

	call	IsThreadWin32
	jc	near .ret

	VMMCall	Get_Sys_VM_Handle
	VMMCall	Get_Initial_Thread_Handle
	mov     ebx,edi

.loop:
	mov	ebp,[edi+TCB_ClientPtr]		; R0TCB.CRS
	test	byte [ebp+CRS.EFlags+2],2	; is client in V86 mode?
	jnz	.next

	call	ShallTraceIt
	jnc	.trace

.next:
	VMMCall	Get_Next_Thread_Handle
	cmp	ebx,edi
	jnz	.loop
	jmp	short .ret

.trace:
	xchg	edi,esi
	mov	ebp,[edi+TCB_ClientPtr]
	mov	ebx,[edi+TCB_VMHandle]

	cmp	byte [OT.OBreaknew],'D'
	jz	@F

	call	BreakIn.ignore_range
	jmp	short .ret

@@
	call	TracerAlloc
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: ThreadInit: failed to allocate TraceInfo, R0TCB: #edi"
	debug_end

	jmp	short .ret

@@
	push	edi
	mov	esi,[ecx+esi]
	mov	edi,[ecx+edi]
	mov	ecx,TraceInfo_size
	rep	movsb
	pop	edi

	mov	esi,[TDS]
	mov	esi,[esi+edi]
	call	TrReinforce

.ret:
	pop	ebp
	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	ebx
	clc
	retn


;-------------------------------------------------------------------------------
; decide if the new thread should be traced or not
;
; ESI: R0TCB of new thread, cannot be in V86 mode
; EDI: R0TCB of an already existing thread
;
; clc: trace it
;-------------------------------------------------------------------------------
ShallTraceIt:
	push	ebx
	push	ecx
	push	edx

	cmp	byte [OT.OTracethread],'D'
	jz	.check_parent

	mov	cx,[esi+TCB_PMPSPSelector]
	cmp	[edi+TCB_PMPSPSelector],cx	; same process?
	jnz	.check_parent

	mov	ecx,[TDS]
	cmp	dword [edi+ecx],byte 0		; already tracing?
	jz	near .no_trace

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: detected new thread in process, R0TCB: #esi"
	debug_end

	jmp	short .trace

.check_parent:
	cmp	byte [OT.OTraceprocess],'D'
	jz	.no_trace

	mov	ebx,[esi+TCB_VMHandle]
	mov	eax,[esi+TCB_ClientPtr]
	movzx	eax,word [eax+CRS.FS]
	VMMCall	_SelectorMapFlat, ebx, eax, byte 0

	cmp	eax,0x80000000
	jb	.no_trace

	cmp	eax,0xC0000000
	jae	.no_trace

	mov	eax,[eax+0x30]			; PDB
	push	dword [eax+0x48]		; parent PDB

	mov	ebx,[edi+TCB_VMHandle]
	mov	eax,[edi+TCB_ClientPtr]
	movzx	eax,word [eax+CRS.FS]
	VMMCall	_SelectorMapFlat, ebx, eax, byte 0
	pop	ecx

	cmp	eax,0x80000000
	jb	.no_trace

	cmp	eax,0xC0000000
	jae	.no_trace

	mov	eax,[eax+0x30]			; PDB
	cmp	eax,ecx				; child/parent?
	jnz	.no_trace

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: detected new thread in child process, R0TCB: #esi"
	debug_end

.trace:
	pop	edx
	pop	ecx
	pop	ebx
	clc
	retn

.no_trace:
	pop	edx
	pop	ecx
	pop	ebx
	stc
	retn


;-------------------------------------------------------------------------------
; stop tracing thread
;
; EDI: R0TCB
;-------------------------------------------------------------------------------
TracerThreadNotExecuteable:
	mov     eax,[TDS]
	mov	eax,[edi+eax]		; TraceInfo
	cmp	eax,byte 0
	jz	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	push	edx
	mov	edx,[eax+TraceInfo.TickCount+4]
	mov	eax,[eax+TraceInfo.TickCount]
	Trace_Out "ICEDUMP: thread died, instruction count: #edx:#eax, R0TCB: #edi"
	pop	edx
	debug_end

	call	TracerFree

@@
	clc
	retn


;-------------------------------------------------------------------------------
; alloc tracer's structure
;
; EDI: R0TCB
;
; stc on error
;-------------------------------------------------------------------------------
TracerAlloc:
	mov     eax,[TDS]
	mov	eax,[edi+eax]
	or	eax,eax
	jz	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: thread is already being traced, R0TCB: #edi"
	debug_end

	stc
	retn

@@
	push	ecx
	push	edx
	VMMCall	_HeapAllocate, dword TraceInfo_size, byte HEAPZEROINIT
	pop	edx
	pop	ecx
	or	eax,eax
	jnz	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: failed to allocate TraceInfo for thread, R0TCB: #edi"
	debug_end

	stc
	retn

@@
	push	eax
	mov     eax,[TDS]
	add	eax,edi
	pop	dword [eax]

; set up fake IDT: int32 gates to a simple iretd
	push	ebx
	push	ecx
	push	edx

	mov	ecx,0x60
	mov	eax,[eax]		; TraceInfo
	lea	edx,[eax+TraceInfo.IRETD]
	mov	ebx,0x00280000
	mov	bx,dx
	mov	dx,0x8E00		; int32, dpl=0
	
@@
	mov	[eax+8*ecx-8+TraceInfo.IDT],ebx
	mov	[eax+8*ecx-8+TraceInfo.IDT+4],edx
	loop	@B

	mov	byte [eax+TraceInfo.IRETD],0xCF
	and	dword [eax+TraceInfo.TickCount],byte 0

	pop	edx
	pop	ecx
	pop	ebx

	clc
	retn


;-------------------------------------------------------------------------------
; free tracer's structure if thread was traced
;
; EDI: R0TCB
;-------------------------------------------------------------------------------
TracerFree:
	push	ecx
	push	edx

	mov     ecx,[TDS]
	xor	eax,eax
	xchg	eax,[edi+ecx]
	or	eax,eax
	jz	@F

	VMMCall	_HeapFree, eax, byte 0

@@
	pop	edx
	pop	ecx
	clc
	retn


;-------------------------------------------------------------------------------
; try to set EFLAGS.T in thread's context
;
; EDI: R0TCB
;-------------------------------------------------------------------------------
StartTracing:
	push	ebp

	mov	ebp,[edi+TCB_ClientPtr]
	bts	word [ebp+CRS.EFlags],8		; set single step flag
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: StartTracing: EFLAGS.T was already set, R0TCB: #edi"
	debug_end

@@
	pop	ebp
	clc
	retn


;-------------------------------------------------------------------------------
; try to reset EFLAGS.T in thread's context
;
; EDI: R0TCB
;-------------------------------------------------------------------------------
StopTracing:
	push	ebp

	mov	ebp,[edi+TCB_ClientPtr]
	btr	word [ebp+CRS.EFlags],8		; reset single step flag
	jc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: StopTracing: EFLAGS.T was not set, R0TCB: #edi"
	debug_end

@@
	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: stopped tracing thread, R0TCB: #edi"
	debug_end

	pop	ebp
	clc
	retn


;-------------------------------------------------------------------------------
; eax: client EIP, skip over known prefixes
;
; eax: first non-prefix byte (real opcode)
; ecx: max length for real opcode (considering the 15 byte limit)
;-------------------------------------------------------------------------------
SkipPrefixes:
	push	esi
	mov	esi,eax

	mov	ecx,15		; max instruction length

.loop:

.protect_start:
	movzx	eax,byte [esi]
.protect_end:

	bt	[Prefixes],eax
	jc	@F

	mov	eax,esi
	pop	esi
	clc
	retn

@@
	inc	esi
	loop	.loop

.EH:
	mov	eax,esi
	pop	esi
	stc
	retn


;segment _LDATA
; prefixes:	db 0x26,0x2E,0x36,0x3E,0x64,0x65,0x66,0x67,0xF0,0xF2,0xF3
	align 4
Prefixes: ;11111111111111110000000000000000
	  ;FEDCBA9876543210FEDCBA9876543210
	dd 00000000000000000000000000000000b ; 1F-00
	dd 01000000010000000100000001000000b ; 3F-20
	dd 00000000000000000000000000000000b ; 5F-40
	dd 00000000000000000000000011110000b ; 7F-60
	dd 00000000000000000000000000000000b ; 9F-80
	dd 00000000000000000000000000000000b ; BF-A0
	dd 00000000000000000000000000000000b ; DF-C0
	dd 00000000000011010000000000000000b ; FF-E0
	dd 0


segment _LTEXT

;-------------------------------------------------------------------------------
; TRACEX <EIP low> [<EIP high>]
;-------------------------------------------------------------------------------
Parse_TraceX:
	mov	edi,Error_V86
	mov	ebp,[dClient_EFLAGS]
	test	byte [ebp+2],2			; is client in V86 mode?
	jnz	near Parser.errorMsg

	mov	edi,Error_PM16
	mov	ebp,[dClient_CS]
	lar	eax,[ebp]			; is client 32 bit?
	bt	eax,22
	jnc	near Parser.errorMsg

	mov	edi,Error_BadEIPlow
	call	ParseExpression			; parse <EIP low>
	jb	near Parser.errorMsg

	mov	[.EIPlow],eax

	call	[pSkipWhiteSpace]		; skip to <EIP high>
	mov	eax,[.EIPlow]
	jz	@F

	mov	edi,Error_BadEIPhigh
	call	ParseExpression			; parse <EIP high>
	jb	near Parser.errorMsg

@@
	mov	[.EIPhigh],eax

; save client EAX/CS/EIP
	mov	ebp,[dClient_EAX]
	push	dword [ebp]
	pop	dword [.EAX]

	mov	ebp,[dClient_CS]
	push	dword [ebp]
	pop	dword [.CS]

	mov	ebp,[dClient_EIP]
	push	dword [ebp]
	pop	dword [.EIP]

; set up registers for service
	push	byte SERVICE_TRACEX
	mov	ebp,[dClient_EAX]
	pop	dword [ebp]

	call	SetCB
	jc	near Parser.error

	mov	ebp,[fExecuteMoreCommands]	; set internal Winice flag to 0
	mov	byte [ebp],0

	popad
	retn


segment _LDATA
	align 4
.EAX:		dd 0
.CS:		dd 0
.EIP:		dd 0
.EIPlow:	dd 0
.EIPhigh:	dd 0


segment _LTEXT

Service_TraceX:
	VMMCall	Get_Cur_VM_Handle
	VMMCall	Get_Cur_Thread_Handle
	call	TracerFree
	call	TracerAlloc
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: TRACEX failed to allocate TraceInfo"
	debug_end

	jmp	short .ret

@@
	mov	esi,[TDS]
	mov	esi,[esi+edi]			; TraceInfo

	push	dword [Parse_TraceX.EIPlow]
	pop	dword [esi+TraceInfo.EIPlow]

	push	dword [Parse_TraceX.EIPhigh]
	pop	dword [esi+TraceInfo.EIPhigh]

	push	dword [Parse_TraceX.EAX]
	pop	dword [ebp+CRS.EAX]

	push	dword [Parse_TraceX.CS]
	pop	dword [ebp+CRS.CS]

	push	dword [Parse_TraceX.EIP]
	pop	dword [ebp+CRS.EIP]

	call	TrReinforce

.ret:
	popfd
	popad
	retn


;-------------------------------------------------------------------------------
; TRACE [<R0TCB> [<EIP low> [<EIP high>]]]
;-------------------------------------------------------------------------------
Parse_Trace:
	push	byte SERVICE_TRACE
	mov	ebp,[dClient_EAX]
	pop	dword [ebp]

	and	dword [.R0TCB],byte 0
	or	dword [.EIPlow],byte -1
	and	dword [.EIPhigh],byte 0

	call	ParseExpression			; parse <R0TCB>
	jc	.setCB

	mov	[.R0TCB],eax

	call	[pSkipWhiteSpace]		; skip to <EIP low>
	jz	.setCB

	mov	edi,Error_BadEIPlow
	call	ParseExpression			; parse <EIP low>
	jb	near Parser.errorMsg

	mov	[.EIPlow],eax

	call	[pSkipWhiteSpace]		; skip to <EIP high>
	mov	eax,[.EIPlow]
	jz	@F

	mov	edi,Error_BadEIPhigh
	call	ParseExpression			; parse <EIP high>
	jb	near Parser.errorMsg

@@
	mov	[.EIPhigh],eax

.setCB:
	call	SetCB
	jc	near Parser.error

	xor	eax,eax
	inc	eax

	mov	ebp,[fPAGEIN_InProgress]	; set internal Winice flag to 1
	mov	[ebp],eax
	mov	ebp,[fExecuteMoreCommands]	; set internal Winice flag to 0
	mov	[ebp],ah

	popad
	retn


segment _LDATA
	align 4
.R0TCB:		dd 0
.EIPlow:	dd 0
.EIPhigh:	dd 0


segment _LDATA
Error_BadEIPlow:	db 'invalid EIP low',0
Error_BadEIPhigh:	db 'invalid EIP high',0


segment _LTEXT
;-------------------------------------------------------------------------------
; prepare for tracing the specified thread, take care of self-tracing guys too
;-------------------------------------------------------------------------------
Service_Trace:
	cmp	[Parse_Trace.R0TCB],byte 0		; print trace info?
	jnz	.trace

	Trace_Out "ICEDUMP: threads under trace:"

	VMMCall	Begin_Critical_Section

	VMMCall	Get_Sys_VM_Handle
	VMMCall	Get_Initial_Thread_Handle
	mov     ebx,edi
	mov     ecx,[TDS]

@@
	cmp	[edi+ecx],byte 0

	Trace_OutNZ "#edi"

	VMMCall	Get_Next_Thread_Handle
	cmp	ebx,edi
	jnz	@B

	VMMCall	End_Critical_Section

.ret1:
	popfd
	popad
	retn

.trace:
	VMMCall	Get_Cur_VM_Handle
	VMMCall	Get_Cur_Thread_Handle

	cmp	[Parse_Trace.R0TCB],edi		; trace current thread?
	jnz	near .trace_notcurrent

	mov	esi,[TDS]
	mov	esi,[esi+edi]			; TraceInfo
	cmp	esi,byte 0			; tracing current thread?
	jnz	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: use TRACEX to trace the current thread"
	debug_end

	jmp	short .ret1

@@
	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: you shot yourself in the foot, use TRACEX to reactivate tracing the current thread"
	debug_end

	call	TracerFree
	jmp	short .ret1

.trace_notcurrent:
	mov	edi,[Parse_Trace.R0TCB]
	VMMCall	Validate_Thread_Handle
	jnc	@F

	Trace_Out "ICEDUMP: invalid thread handle, R0TCB: #edi"

	jmp	.ret2

@@
	mov	esi,[TDS]
	mov	esi,[esi+edi]

	mov	ecx,[Parse_Trace.EIPlow]
	mov	edx,[Parse_Trace.EIPhigh]
	cmp	ecx,edx
	jbe	.start

	cmp	esi,byte 0
	jnz	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: thread was not traced, R0TCB: #edi"
	debug_end

	jmp	short .ret2

@@
	cmp	dword [esi+TraceInfo.State0],byte ST0_SELFTRACEON
	jz	@F

	call	StopTracing

@@
	call	TracerFree
	jmp	short .ret2

.start:
	cmp	esi,byte 0			; already tracing?
	jnz	.update_range

	mov	ebp,[edi+TCB_ClientPtr]
	test	byte [ebp+CRS.EFlags+2],2	; is client in V86 mode?
	jz	@F

;	call	IsThreadWin32
;	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: thread is in V86 mode, R0TCB: #edi"
	debug_end

	jmp	short .ret2

@@
	call	TracerAlloc
	jnc	@F

	debug_start sdata+DebugFlags, ICEDUMP_DEBUG_TRACE
	Trace_Out "ICEDUMP: TRACE failed to allocate TraceInfo, R0TCB: #edi"
	debug_end

	jmp	short .ret2

@@
	mov	esi,[TDS]
	mov	esi,[esi+edi]
	call	TrReinforce

.update_range:
	mov	dword [esi+TraceInfo.EIPlow],ecx
	mov	dword [esi+TraceInfo.EIPhigh],edx

.ret2:
	popfd
	popad
	retn

%endif
